package notio;

import java.util.*;
import java.io.Serializable;

    /**
     * This class provides an adjacency matrix representation suitable for conceptual
     * graphs.  The matrix is three-dimensional.  Since CG's are bipartite, 
     * the first and second dimensions correspond to relations
     * and concepts respectively.  The first element in the third dimension is the
     * number of arcs between the specified concept and relation.  Note that CG's can 
     * indeed have multiple arcs between a concept and relation since arcs are ordered
     * and therefore distinct.  A zero indicates that the nodes are not adjacent.
     * The remaining elements along the third dimension of an entry are the numbers of the
     * arcs connecting the two nodes.
     * For example, if a relation and concept had two arcs number 1 and 4 connecting them,
     * the entry would be an array like this: { 2, 1, 4 }.
     * Of course, if one only wishes to test the adjacency of two nodes, one can simply
     * look to see if the first element of the third dimension is non-zero.
     * Once generated, instances of this class are independent of the graph from which they
     * are derived.  That is, if the original graph changes, the adjacency matrix is not
     * updated.
     * 
     * This alternate representation is provided to allow for the easy implementation of
     * certain graph algorithms.  
     * Since efficiency is a primary concern and instances of this class are not updated
     * along with the structures upon which they are based, the usual Notio convention of
     * using accessor (getX/putX methods) which provide safe access to the information is 
     * avoided and public members are used instead.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.3 $, $Date: 2000/08/28 20:58:27 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
     */
public class AdjacencyMatrix implements Serializable
  {
  	/** The array of concepts indicating the concept ordering. **/
  public Concept concepts[];

  	/** The array of relations indicating the relation ordering. **/
  public Relation relations[];

  	/** A two dimensional array with the adjacencies.  The first index is for relations,
  	 * the second for concepts, and the third is for arc count followed by arc numbers. **/
  public int adjacencies[][][];

  	/**
  	 * Constructs an AdjacencyMatrix with no initial values for the fields.
  	 */
  public AdjacencyMatrix()
  	{
  	}  	 
  
  	/**
  	 * Constructs an AdjacencyMatrix by copying the supplied arrays.
  	 *
  	 * @param newConcepts  the concept ordering array for the matrix.
  	 * @param newRelations  the relation ordering array for the matrix.
  	 * @param newAdjacencies  the matrix of adjancies.
  	 */
  public AdjacencyMatrix(Concept newConcepts[], Relation newRelations[], int newAdjacencies[][][])
  	{
  	concepts = new Concept[newConcepts.length];
  	System.arraycopy(concepts, 0, newConcepts, 0, newConcepts.length);

  	relations = new Relation[newRelations.length];
  	System.arraycopy(relations, 0, newRelations, 0, newRelations.length);
  	
  	adjacencies = new int[newRelations.length][newConcepts.length][];
  	for (int rel = 0; rel < newRelations.length; rel++)
  		for (int con = 0; con < newConcepts.length; con++)
  			{
  			adjacencies[rel][con] = new int [newAdjacencies[rel][con].length];
	  		System.arraycopy(newAdjacencies[rel][con], 0, adjacencies[rel][con], 0, newAdjacencies[rel][con].length);	  		
	  		}
  	}
  }
