package notio;

import java.util.Hashtable;
import java.io.Serializable;

    /** 
     * Class for designators with a descriptor (Graph) designator.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.13 $, $Date: 1999/05/04 01:35:50 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
     */

public class DescriptorDesignator extends Designator implements Serializable
  {
  /** Graph descriptor. **/
  private Graph descriptor;
  /** Marker associated with descriptor graph. **/
  private Marker marker;

    /**
     * Constructs a new DescriptorDesignator with the specified graph as a
     * descriptor.
     *
     * @param newDescriptor  the descriptor graph.
     * @param markerSet  the marker set used in conjunction with this
     * designator.
     */
  public DescriptorDesignator(Graph newDescriptor, MarkerSet markerSet)
    {
    descriptor = newDescriptor;

    // Lookup marker or create one if necessary
    marker = markerSet.getMarkerByIndividual(descriptor);
    if (marker == null)
      marker = new Marker(markerSet, descriptor);
      
    descriptor.setEnclosingDesignator(this);
    }

    /**
     * Returns a constant indicating which kind of designator this is.
     * In this case the constant will be: Designator.DESIGNATOR_DESCRIPTOR
     *
     * @return a constant indicating the kind of the designator.
     */
  public int getDesignatorKind()
    {
    return DESIGNATOR_DESCRIPTOR;
    }

    /**
     * Returns the Graph descriptor.
     *
     * @return the Graph descriptor.
     */
  public Graph getDescriptor()
    {
    return descriptor;
    }

    /**
     * Returns the marker associated with this descriptor.  Note that this
     * is equivalent to a call to MarkerSet.getMarkerByIndividual() with the
     * descriptor Graph as an argument.  This means that the marker is
     * associated with the actual Graph object.  As such, the marker is
     * probably not a good basis for comparison of descriptors.  Some form of
     * graph matching should be used instead.
     *
     * @return the marker associated with the describing graph.
     */
  public Marker getMarker()
    {
    return marker;
    }

  	/**
     * Performs a copy operation on this designator according to the
     * the specified CopyingScheme.
     * The result may be a new designator or simply a reference to this designator
     * depending on the scheme.
  	 *
  	 * @param copyScheme  the copying scheme used to control the copy operation.
     * @param substitutionTable  a hashtable containing copied objects available due to 
     * earlier copy operations.
  	 * @return the result of the copy operation.
  	 */
  public Designator copy(CopyingScheme copyScheme, Hashtable substitutionTable)
  	{
  	switch (copyScheme.getDesignatorFlag())
  		{
  		case CopyingScheme.DG_COPY_DUPLICATE:
  			{
  			switch (copyScheme.getGraphFlag())
  				{
  				case CopyingScheme.GR_COPY_DUPLICATE:
  					{
  					CopyingScheme nestedScheme;
  					
  					nestedScheme = copyScheme.getNestedCopyingScheme();
  					// If the nested scheme is null we use the current copying scheme.
  					if (nestedScheme == null)
  						nestedScheme = copyScheme;
  						
						return new DescriptorDesignator(getDescriptor().copy(copyScheme, substitutionTable), marker.getMarkerSet());
						}
						
					case CopyingScheme.GR_COPY_REFERENCE:
						{
						return new DescriptorDesignator(getDescriptor(), marker.getMarkerSet());
						}
					}
  			}
  		
  		case CopyingScheme.DG_COPY_REFERENCE:
  			{
  			return this;
  			}
  			
  		default:
        throw new UnimplementedFeatureException("Specified Designator copy control flag is unknown.");    		  		
  		}
  	}
  }
