package notio;

import java.util.*;

    /** 
     * A class for storing the referent of a concept.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.18 $, $Date: 1999/05/04 01:35:58 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
     */
public class Referent implements java.io.Serializable
  {
  /** The set of designators for this referent. **/
  private Designator designators[];
  /** The quantifier for this referent. **/
  private Macro quantifier;
  /** The coreference sets to which this referent belongs. **/
  private CoreferenceSet corefSets[];
  /** The concept that encloses this referent. **/
  private Concept enclosingConcept;
  

  // Let's see:
  // - Referents differentiate graphs
  // - Multiple designators
  // - Coreference (resolution to entity)
  // - Designator
  // - Quantifier
  // - Sets
  // - Ownership of referent
  // - Freeform referent?

		/**
		 * Constructs a new referent with the specified quantifier and array of designators.
		 *
		 * @param newQuantifier  the quantifier for this referent.
		 * @param newDesignators  an array of designators for this referent.
		 */
  public Referent(Macro newQuantifier, Designator newDesignators[])
  	{
  	quantifier = newQuantifier;
  	
  	if (newDesignators != null)
  		{
  		designators = new Designator[newDesignators.length];
  		System.arraycopy(newDesignators, 0, designators, 0, newDesignators.length);

  		// Set this referent as the enclosing referent for each designator
	  	for (int des = 0; des < designators.length; des++)
  				designators[des].setEnclosingReferent(this);
  		}  		
  	}
  
		/**
		 * Constructs a new referent with the specified quantifier and single designator.
		 *
		 * @param newQuantifier  the quantifier for this referent.
		 * @param newDesignator  a single designator for this referent.
		 */
  public Referent(Macro newQuantifier, Designator newDesignator)
  	{
  	quantifier = newQuantifier;
  	
  	if (newDesignator != null)
  		{
  		designators = new Designator[1];
  		designators[0] = newDesignator;
	  	designators[0].setEnclosingReferent(this);
  		}  		
  	}
  
		/**
		 * Constructs a new referent with the specified quantifier and no designator.
		 *
		 * @param newQuantifier  the quantifier for this referent.
		 */
  public Referent(Macro newQuantifier)
  	{
  	quantifier = newQuantifier;
  	}
  	
		/**
		 * Constructs a new referent with no quantifier and an array of designators.
		 *
		 * @param newDesignators  an array of designators for this referent.
		 */
  public Referent(Designator newDesignators[])
  	{
  	quantifier = null;
  	
  	if (newDesignators != null)
  		{
  		designators = new Designator[newDesignators.length];
  		System.arraycopy(newDesignators, 0, designators, 0, newDesignators.length);
  		}
  		
  	// Set this referent as the enclosing referent for each designator
  	for (int des = 0; des < designators.length; des++)
  			designators[des].setEnclosingReferent(this);
  	}
  	
		/**
		 * Constructs a new referent with no quantifier and a single designator.
		 *
		 * @param newDesignator  a single designator for this referent.
		 */
  public Referent(Designator newDesignator)
  	{
  	quantifier = null;
  	
  	if (newDesignator != null)
  		{
  		designators = new Designator[1];
  		designators[0] = newDesignator;
	  	designators[0].setEnclosingReferent(this);
  		}  		
  	}
  
    /**
     * Returns this referent's quantifier (null indicates existential
     * quantifier).
     *
     * @return the concept's quantifier.
     */
  public Macro getQuantifier()
    {
    return quantifier;
    }

    /**
     * Returns this concept's designators (null indicates generic concept).
     *
     * @return an array containing this concept's designators or null.
     */
  public Designator[] getDesignators()
    {
    return designators;
    }
    
		/**
		 * Returns true if this referent forms a context (has a descriptor designator).
		 *
		 * @return true if this referent forms a context.
		 */
	public boolean isContext()
		{
		if (designators == null)
			return false;
			
		for (int des = 0; des < designators.length; des++)
			if (designators[des].getDesignatorKind() == Designator.DESIGNATOR_DESCRIPTOR)
				return true;
				
		return false;
		}

  	/**
  	 * Returns the concept that encloses this referent.
  	 *
  	 * @return the concept that encloses this referent.
  	 */
	public Concept getEnclosingConcept()
		{
		return enclosingConcept;
		}
		
		/**
		 * Sets the enclosing concept for this referent.
		 *
		 * @param newEnclosingConcept  the new enclosing concept for this referent.
		 * @bug Should this throw an exception if it called when already set? Probably not since
		 * it is handled internally.
		 *
     * @impspec this method is present to support the getEnclosingConcept() method.
		 */
	void setEnclosingConcept(Concept newEnclosingConcept)
		{
		enclosingConcept = newEnclosingConcept;
		}
		
    /**
     * Performs a copy operation on this referent according to the
     * the specified CopyingScheme.
     * The result may be a new referent or simply a reference to this referent
     * depending on the scheme.
     * Coreference sets will be copied as required by the CopyingScheme.
     *
     * @param copyScheme  the copying scheme used to control the copy operation.
     * @return the result of the copy operation.
     */
  public Referent copy(CopyingScheme copyScheme)
    {
    return copy(copyScheme, new Hashtable());
    }

    /**
     * Performs a copy operation on this referent according to the
     * the specified CopyingScheme.
     * The result may be a new referent or simply a reference to this referent
     * depending on the scheme.
     * Coreference sets will be copied as required by the CopyingScheme.
     *
     * @param copyScheme  the copying scheme used to control the copy operation.
     * @param substitutionTable  a hashtable containing copied objects available due to
     * earlier copy operations.
     * @return the result of the copy operation.
     * @bug Currently uses CN_ copy flags.  Should probably add referent flags.
     */
  public Referent copy(CopyingScheme copyScheme, Hashtable substitutionTable)
    {
    switch (copyScheme.getConceptFlag())
    	{
    	case CopyingScheme.CN_COPY_DUPLICATE:
    		{
    		Referent newReferent;
				Designator newDesignators[];

				// If this referent is already already in the substitution table, return its
				// substitute.
				newReferent = (Referent)substitutionTable.get(this);
				if (newReferent != null)
					return newReferent;

				// Create the new referent, copying the designators as we do so.
				if (designators != null)
					{
					newDesignators = new Designator[designators.length];
					
					for (int des = 0; des < designators.length; des++)
						newDesignators[des] = designators[des].copy(copyScheme, substitutionTable);
					}
				else
					newDesignators = null;

		    newReferent = new Referent(getQuantifier(), newDesignators);

				// Add pair to substitution table
				substitutionTable.put(this, newReferent);

    		return newReferent;
    		}

    	case CopyingScheme.CN_COPY_REFERENCE:
    		{
    		return this;
    		}

    	default:
        throw new UnimplementedFeatureException("Specified Concept copy control flag is unknown.");
    	}
		}
		
    /**
     * Compares two referents to decide if they match.  The exact semantics of matching
     * are determined by the matching scheme.  
     *
     * @param first  the first referent being matched.
     * @param second  the second referent being matched.
     * @param matchingScheme  the matching scheme that determines how the match is performed.
     * @return true if the two referents match according to the scheme's criteria.
     * @bug Should add specified quantifier matching too.
     * @bug Doesn't really handle multiple designators.  Requires extensive thought.
     */
  public static MatchResult matchReferents(Referent first, Referent second,
    MatchingScheme matchingScheme)
    {
    // If either referent is null then they automatically match
    if ((first == null) || (second == null))
    	return new MatchResult(true);
    	
    Designator firstDesignators[], secondDesignators[];
    
    firstDesignators = first.getDesignators();
    secondDesignators = second.getDesignators();
    
    if ((firstDesignators == null) || (secondDesignators == null))
    	return new MatchResult(true);		    
    	
		return Designator.matchDesignators(firstDesignators[0], secondDesignators[0],
			matchingScheme);
    }
  }
