package notio;

import java.util.Hashtable;
import java.io.Serializable;

    /** 
     * Class for set designators.  A designator of this kind specifies a
     * collective set of other designators.  
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.10 $, $Date: 1999/05/04 01:36:01 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
     */

public class SetDesignator extends Designator implements Serializable
  {
  /** The set of designators associated with this set designator. **/
  private Set designatorSet;
  /** The marker object associated with this set designator. **/
  private Marker marker;

    /**
     * Constructs a new SetDesignator with the specified designators.  If null
     * is passed in place of an array, the designator set is assumed to be
     * generic.
     * // BUG: Currently SetDesignators may contain other SetDesignators.  Is
     * // this a problem or a feature?
     *
     * @param newDesignatorSet  an array of designators that will form the
     * set or null to indicate generic set.
     * @param markerSet  the markerSet associated with this designator.
     */
  public SetDesignator(Designator newDesignatorSet[], MarkerSet markerSet)
    {
    if (newDesignatorSet == null)
      designatorSet = null;
    else
      {
      designatorSet = new Set();
      for (int des = 0; des < newDesignatorSet.length; des++)
        designatorSet.addElement(newDesignatorSet[des]);
      }

    // Lookup marker or create one if necessary.
    marker = markerSet.getMarkerByIndividual(designatorSet);
    if (marker == null)
      marker = new Marker(markerSet, designatorSet);
    }

    /**
     * Returns a constant indicating which kind of designator is.
     * In this case the constant will be: Designator.DESIGNATOR_SET
     *
     * @return a constant indicating the kind of the designator.
     */
  public int getDesignatorKind()
    {
    return DESIGNATOR_SET;
    }

    /**
     * Returns an array of the designators in this set or null, indicating a
     * generic set.
     *
     * @return an array of the designators in this set or null, indicating a
     * generic set.
     */
  public Designator[] getDesignatorSet()
    {
    Designator desArr[];

    if (designatorSet == null)
      return null;

    desArr = new Designator[designatorSet.size()];
    designatorSet.copyInto(desArr);

    return desArr;
    }

    /**
     * Returns the marker associated with the designator set.  This is
     * equivalent to a lookup using MarkerSet.getMarkerByIndividual() with the
     * internal representation of the designator set as an argument.  As such,
     * it is probably not useful for comparing designators.  Instead, the
     * set should be compared element by element.
     *
     * @return the marker associated with the designator set.
     */
  public Marker getMarker()
    {
    return marker;
    }
    
  	/**
     * Performs a copy operation on this designator according to the
     * the specified CopyingScheme.
     * The result may be a new designator or simply a reference to this designator
     * depending on the scheme.
  	 *
  	 * @param copyScheme  the copying scheme used to control the copy operation.
     * @param substitutionTable  a hashtable containing copied objects available due to 
     * earlier copy operations.
  	 * @return the result of the copy operation.
  	 * @bug Should probably copy all designators in the set.  Currently does not.
  	 */
  public Designator copy(CopyingScheme copyScheme, Hashtable substitutionTable)
  	{
  	switch (copyScheme.getDesignatorFlag())
  		{
  		case CopyingScheme.DG_COPY_DUPLICATE:
  			{
  			return new SetDesignator(getDesignatorSet(), getMarker().getMarkerSet());
  			}
  		
  		case CopyingScheme.DG_COPY_REFERENCE:
  			{
  			return this;
  			}
  			
  		default:
        throw new UnimplementedFeatureException("Specified Designator copy control flag is unknown.");    		  		
  		}
  	}
  }
