package notio;

import java.util.*;
import java.io.Serializable;

    /** 
     * A class for holding information related to translation so it can be
     * given to different parsers and generators to ensure consistent
     * representations.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.12 $, $Date: 1999/05/04 01:36:02 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
     * @bug Should we really provide a native marker for foreign marker IDs or should this
     * information be kept purely syntatically?  That would remove the need for this sort
     * of translation lookup table but would render comparison somewhat less efficient.
     * @bug Should we provide a way to remove specific mappings?  Not able to add null mappings
     * at present, but that would be one way to do it.
     */

public class TranslationContext implements Serializable
  {
  /** A lookup table that maps coreference labels to coreference sets. **/
  private Hashtable defLabelToCorefSetTable;
  /** A lookup table that maps coreference sets to coreference labels. **/
  private Hashtable corefSetToDefLabelTable;
  
  /** A lookup table that maps foreign marker ids to native markers. **/
  private Hashtable foreignIDToNativeMarkerTable;
  /** A lookup table that maps native markers to foreign marker ids. **/
  private Hashtable nativeMarkerToForeignIDTable;
  
  /** A factory for producing defining labels. **/
  private LabelFactory defLabelFactory;

    /**
     * Constructs a new translation context.
     */
  public TranslationContext()
    {
    defLabelToCorefSetTable = new Hashtable();
    corefSetToDefLabelTable = new Hashtable();
    foreignIDToNativeMarkerTable = new Hashtable();
    nativeMarkerToForeignIDTable = new Hashtable();
    defLabelFactory = new LabelFactory();
    }

    /**
     * Constructs a new translation context that is a copy of the specified original.
     * Changes to the new context will not affect the old one or vice versa.
     *
     * @param originalContext  the translation context to be copied.
     */
  public TranslationContext(TranslationContext originalContext)
    {
    String lastLabel;
    
    defLabelToCorefSetTable = copyHashtable(originalContext.getDefLabelToCorefSetTable());
    corefSetToDefLabelTable = copyHashtable(originalContext.getCorefSetToDefLabelTable());
    foreignIDToNativeMarkerTable = copyHashtable(originalContext.getForeignIDToNativeMarkerTable());
    nativeMarkerToForeignIDTable = copyHashtable(originalContext.getNativeMarkerToForeignIDTable());
    lastLabel = originalContext.getLastDefiningLabel();
    
    if (lastLabel == null)
    	defLabelFactory = new LabelFactory();
    else
    	defLabelFactory = new LabelFactory(lastLabel);
		}

    /**
     * Creates a two-way mapping between a defining label and a coreference set instance.
     * This table may be used by
     * parsers to associate coreference sets with defining labels.  This is
     * particularly necessary for parsing or generating nested graphs with 
     * lines of identity.  It is difficult for a parser or generator to
     * intelligently decide when this mapping is no longer valid (e.g. when
     * parsing a new graph that uses the same defining labels) so it may be
     * cleared by the user or controlling application using the
     * clearDefiningLabelToCoreferenceSetMapping() method.
     *
     * @param definingLabel  the defining label being mapped.
     * @param corefSet  the coreference set being mapped.
     */
  public void mapDefiningLabelToCoreferenceSet(String definingLabel, CoreferenceSet corefSet)
    {
    defLabelToCorefSetTable.put(definingLabel, corefSet);
    corefSetToDefLabelTable.put(corefSet, definingLabel);
    }
    
    /**
     * Returns the coreference set associated with the specified defining label in this 
     * translation context.
     *
     * @param definingLabel  the defining label used for lookup.
     * @return the coreference set associated with definingLabel, or null if no mapping exists.
     */
  public CoreferenceSet getCoreferenceSetByDefiningLabel(String definingLabel)
    {
    return (CoreferenceSet)defLabelToCorefSetTable.get(definingLabel);
    }

    /**
     * Returns the defining label associated with the specified coreference set in this 
     * translation context.
     *
     * @param corefSet  the coreference set used for lookup.
     * @return the defining label associated with the set, or null if no mapping exists.
     */
  public String getDefiningLabelByCoreferenceSet(CoreferenceSet corefSet)
    {
    return (String)corefSetToDefLabelTable.get(corefSet);
    }
    
    /**
     * Clears all existing mappings between coreference labels and coreference sets.
     */
  public void clearDefiningLabelToCoreferenceSetMapping()
    {
    defLabelToCorefSetTable.clear();
    corefSetToDefLabelTable.clear();
    }

    
    /**
     * Creates a two-way mapping between a foreign marker ID and a native marker instance.
     * This is used to facilitate marker assignment and lookup while translating.  During 
     * parsing, newly encountered foreign marker ID's can have a corresponding native marker
     * instantiated and a mapping created.  Subsequent references to the foreign ID during
     * during parsing can then use the same native marker.  The table may be used for 
     * generation purposes in order to preserve consistency between parsed and generated
     * markers.
     * This mapping may need to be cleared in some situations.  For example, when parsing two
     * graphs in succession that contain identical foreign marker ID's but whose resulting
     * markers should not be identical internally.  Exactly when the table should be cleared
     * should be decided by the application.
     *
     * @param foreignID  the foreign marker ID being mapped.
     * @param marker  the native marker being mapped.
     */
  public void mapForeignMarkerIDToNativeMarker(String foreignID, Marker marker)
    {
    foreignIDToNativeMarkerTable.put(foreignID, marker);
    nativeMarkerToForeignIDTable.put(marker, foreignID);
    }
    
    /**
     * Returns the native marker associated with the specified foreign marker ID in this 
     * translation context.
     *
     * @param foreignID  the foreign marker ID used for lookup.
     * @return the native marker associated with foreignID, or null if no mapping exists.
     */
  public Marker getNativeMarkerByForeignMarkerID(String foreignID)
    {
    return (Marker)foreignIDToNativeMarkerTable.get(foreignID);
    }

    /**
     * Returns the foreign marker ID associated with the specified native marker in this 
     * translation context.
     *
     * @param marker  the native marker used for lookup.
     * @return the foreign marker ID associated with the marker, or null if no mapping exists.
     */
  public String getForeignMarkerIDByNativeMarker(Marker marker)
    {
    return (String)nativeMarkerToForeignIDTable.get(marker);
    }
    
    /**
     * Clears all existing mappings between foreign marker ID's and native markers.
     */
  public void clearForeignMarkerIDToNativeMarkerMapping()
    {
    foreignIDToNativeMarkerTable.clear();
    nativeMarkerToForeignIDTable.clear();
    }
    
    /**
     * Returns a new defining label that is not already present in the defining 
     * label/coreference set mapping.  There are no constraints on the form of the
     * label other than that they conform to a valid CGIF identifier.  Otherwise,
     * labels' forms are determined by the underlying implementation.
     *
     * @return a new, unique, and unused defining label.
     */
  public String getNextAvailableDefiningLabel()
    {
    String newLabel;
    
    do
    	{
    	newLabel = defLabelFactory.getNextLabel();
    	}
    while (defLabelToCorefSetTable.get(newLabel) != null);
      
    return newLabel;
    }
    
    /**
     * Resets the generator for defining labels so that it starts again.
     * This method would typically be called after clearing the defining label - 
     * coreference set mapping so that labels may be reused in a new translation, 
     * if possible.
     * Implementations are not required to honour this although they should silently
     * do nothing rather than throwing an exception.  If possible, implementations
     * should reset their label generators so that they produce a sequence of
     * labels as if this were a newly constructed context.
     */
  public void resetAvailableDefiningLabel()
    {
    defLabelFactory = new LabelFactory();
    }
    
    /**
     * Returns the defining label to coreference set table as a hashtable.
     *
     * @return the defining label to coreference set table as a hashtable.
     */
  Hashtable getDefLabelToCorefSetTable()
  	{
  	return defLabelToCorefSetTable;
  	}

    /**
     * Returns the coreference set to defining label table as a hashtable.
     *
     * @return the coreference set to defining label table as a hashtable.
     */
  Hashtable getCorefSetToDefLabelTable()
  	{
  	return corefSetToDefLabelTable;
  	}

    /**
     * Returns the foreign ID to native marker table as a hashtable.
     *
     * @return the foreign ID to native marker table as a hashtable.
     */
  Hashtable getForeignIDToNativeMarkerTable()
  	{
  	return foreignIDToNativeMarkerTable;
  	}

    /**
     * Returns the native marker to foreign ID table as a hashtable.
     *
     * @return the native marker to foreign ID table as a hashtable.
     */
  Hashtable getNativeMarkerToForeignIDTable()
  	{
  	return nativeMarkerToForeignIDTable;
  	}

    /**
     * Returns the last defining label generated by this context or null if no labels
     * have generated.
     * Note that this method does not generated any new labels.
     *
     * @return the last defining label generated by this context or null.
     */
  public String getLastDefiningLabel()
  	{
  	return defLabelFactory.getLastLabel();
  	}
  	
		/**
		 * Creates a deep copy of the specified hashtable.
		 *
		 * @param inTable  the hashtable to be copied.
		 * @return a copy of the hashtable.
		 */
	private Hashtable copyHashtable(Hashtable inTable)
		{
		Enumeration keys, elements;
		Hashtable newTable = new Hashtable();
		
		keys = inTable.keys();
		elements = inTable.elements();
		
		while (keys.hasMoreElements())
			newTable.put(keys.nextElement(), elements.nextElement());
			
		return newTable;
		}
  }
