package notio.examples;

import java.io.*;
import java.util.*;
import notio.*;
import notio.translators.*;

    /** 
     * An example of how to use the CGIFParser and CGIFTranslator
     * for parsing and generating CGIF files.  The application reads
     * a CGStream from from a file and then immediately outputs the graphs
     * it read to a different file.
		 *
		 * This example is almost identical to CGIFTranslation, except that it
		 * works with CGStreams instead of individual graphs.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.1 $, $Date: 2001/07/03 11:59:23 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
     */
public class CGStreamTranslation
  {
  	/**
  	 * An application main() method for this example.  This method creates
  	 * the basic structures required for parsing and generating and then calls
  	 * two methods, one to parse a graph from a file, the other to write the
  	 * parsed graph back out to a new file.
  	 *
  	 * @param args  an array of String containing the command line arguments.
  	 */
  public static void main(String args[])
  	{
  	KnowledgeBase kBase;
  	TranslationContext tContext;
  	Vector graphs;
		String inputFilename, outputFilename;

		// READING COMMAND LINE ARGUMENTS
		// First we check for command line arguments which will give us the input
		// and output filenames.  If these are omitted, we will use standard input
		// and standard output.
		if (args.length >= 1)
			inputFilename = args[0];
		else
			inputFilename = null;

		if (args.length >= 2)
			outputFilename = args[1];
		else
			outputFilename = null;

		if (args.length > 2)
			{
			System.err.println("Too many arguments.");
			System.err.println("Usage: java notio.examples.CGIFTranslation [input filename [output filename]]");
			System.exit(1);
			}
  	
  	// CREATING A KNOWLEDGE BASE
  	// The current CGIF Parser and CGIF Translator both require a
  	// KnowledgeBase instance when they are constructed.  The base is
  	// simply a means for passing in the type hierarchies and marker set.
  	kBase = new KnowledgeBase();

		// CREATING A TRANSLATION CONTEXT
		// A translation context is used for the storage of temporary tables
		// during parsing and generating.  Using the same translation context
		// for parsing and generating means that user provided identifiers
		// can be preserved.  For example, the same coreference labels will
		// be used in generating as were found during parsing.  A translation
		// context can be partially or completely cleared if desired.
		tContext = new TranslationContext();

		// We now call a method, shown below, which will parse a graph from 
		// the specified file, or from standard input.  
		if (inputFilename == null)
			graphs = parseCGStreamFromReader(kBase, tContext, 
				new BufferedReader(new InputStreamReader(System.in)));
		else
			graphs = parseCGStreamFromFile(kBase, tContext, inputFilename);

		if (graphs == null)
			System.exit(1);
		
		// This method will output our newly parsed graph into another file, or
		// to standard output.
		if (outputFilename == null)
			{
			PrintWriter writer = new PrintWriter(System.out);
			generateCGStreamIntoWriter(kBase, tContext, 
				writer, graphs);
			writer.flush();
			System.out.println();
			}
		else
			generateCGStreamIntoFile(kBase, tContext, outputFilename, graphs);
  	}

  	/**
  	 * Attempts to parse a graph from the specified Reader.
  	 *
  	 * @param kBase  the knowledge base that will be used during parsing.
  	 * @param tContext  the translation context that will be used during parsing.
  	 * @param reader  the Reader from which to parse.
  	 * @return the parsed graph or null if the parsing fails for any reason.
  	 */
  public static Vector parseCGStreamFromReader(KnowledgeBase kBase, TranslationContext tContext,
  	Reader reader)
  	{
  	CGIFParser parser;
  	Vector graphs = null;
  	
		// CREATING A PARSER
		// We now have all we need to create a CGIFParser instance.
		parser = new CGIFParser();
		try
			{
			parser.initializeParser(reader, kBase, tContext);
			}
	  catch (ParserException e)
	  	{
	  	e.printStackTrace();
	  	System.exit(1);
	  	}

		// PARSING A CGSTREAM FROM THE READER
		try
			{
			// Here, we ask the parser to try to parse a CGStream.  The CGIF parser
			// actually allows any non-terminal of the grammar to be parsed so
			// we could have asked for a Concept, Relation, or any other structure
			// instead.
			graphs = parser.parseCGStream();
			}
		catch (notio.ParserException e)
			{
			// This is the proper Notio exception that all parsers should throw
			// when they encounter problems.
			
			// Report error to user.
			System.err.println("Error while parsing:");
			System.err.println(e.getMessage());
			return null;
			}
		
		return graphs;
  	}

  	/**
  	 * Opens a file for reading and then attempts to parse a graph from it.
  	 *
  	 * @param kBase  the knowledge base that will be used during parsing.
  	 * @param tContext  the translation context that will be used during parsing.
  	 * @param filename  the name of the file to be opened.
  	 * @return the parsed graph or null if the parsing fails for any reason.
  	 */
  public static Vector parseCGStreamFromFile(KnowledgeBase kBase, TranslationContext tContext,
  	String filename)
  	{
  	FileReader reader = null;
  	CGIFParser parser;
  	Vector graphs = null;
  	
		// INITIALIZING A READER
		// Any Reader can be used with the CGIFParser.  In this case we 
		// use a FileReader to read from the file specified by filename.
		try
			{
			reader = new FileReader(filename);
			}
		catch (FileNotFoundException e)
			{
			// Report error to user.
			return null;
			}

		// PERFORM THE ACTUAL PARSE
		graphs = parseCGStreamFromReader(kBase, tContext, reader);

		// CLOSING THE READER
		try
			{
			reader.close();
			}
		catch (IOException e)
			{
			// Report error to user.
			return null;
			}
			
		return graphs;
  	}

  	/**
  	 * Attempts to generate a graph into the specified Writer.
  	 *
  	 * @param kBase  the knowledge base that will be used during generating.
  	 * @param tContext  the translation context that will be used during generating.
  	 * @param writer  the Writer to be written to.
  	 * @param graph  the graph to be generated into the file.
  	 */
  public static void generateCGStreamIntoWriter(KnowledgeBase kBase, TranslationContext tContext,
  	Writer writer, Vector graphs)
  	{
  	CGIFGenerator generator;
  	
		// CREATING A GENERATOR
		// We now have all we need to create a CGIFGenerator instance.
		generator = new CGIFGenerator();
		try
			{
			generator.initializeGenerator(writer, kBase, tContext);
			}
		catch (GeneratorException e)
			{
			// Report error to user.
			System.err.println("Error while generating:");
			System.err.println(e.getMessage());
			return;
			}

		// SET SOME OPTIONS ON THE GENERATOR
		// The following options control how the generated output is formatted.
		// Consult the documentation for CGIFGenerator for details on how they
		// work.
		generator.setEmbedSingleRelatorConcepts(true);
		generator.setInitiateGraphsWithNewline(true);
		generator.setTerminateGraphsWithNewline(true);
		generator.setIndentNestedGraphs(true);

		// GENERATING A CGSTREAM INTO THE WRITER
		try
			{
			// Here, we ask the generator to try and generate the specified graphs
			// into the writer.  The generator actually allows any non-terminal of 
			// the grammar to be generated so we could have generated a Concept, Relation, 
			// or any other structure instead.
			generator.generateCGStream(graphs);
			}
		catch (notio.GeneratorException e)
			{
			// This is the proper Notio exception that all generators should throw
			// when they encounter problems.
			
			// Report error to user.
			return;
			}
  	}
		
  	/**
  	 * Opens a file for writing and then attempts to generate a graph into it.
  	 *
  	 * @param kBase  the knowledge base that will be used during generating.
  	 * @param tContext  the translation context that will be used during generating.
  	 * @param filename  the name of the file to be written to.
  	 * @param graph  the graph to be generated into the file.
  	 */
  public static void generateCGStreamIntoFile(KnowledgeBase kBase, TranslationContext tContext,
  	String filename, Vector graphs)
  	{
  	FileWriter writer = null;
  	CGIFGenerator generator;
  	
		// INITIALIZING A WRITER
		// Any Writer can be used with the CGIFGenerator.  In this case we 
		// use a FileWriter to write to the file specified by filename.
		try
			{
			writer = new FileWriter(filename);
			}
		catch (IOException e)
			{
			// Report error to user.
			return;
			}
			
		// CREATING A GENERATOR
		// We now have all we need to create a CGIFGenerator instance.
		generator = new CGIFGenerator();
		try
			{
			generator.initializeGenerator(writer, kBase, tContext);
			}
		catch (GeneratorException e)
			{
			// Report error to user.
			return;
			}

		// PERFORM THE ACTUAL GENERATION
		generateCGStreamIntoWriter(kBase, tContext, writer, graphs);

		// CLOSE THE WRITER
		try
			{
			writer.close();
			}
		catch (IOException e)
			{
			// Report error to user.
			return;
			}
  	}
  }
