package notio.translators;

import java.util.*;
import notio.*;

    /** 
     * A CGIF Generator class.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.76 $, $Date: 2001/07/03 11:57:38 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
		 *
     * @bug Should add 'useNegationAbbreviation' flag to generator.
		 * @idea Improve formatting by allowing newlines/indentation before/after each node.
     */
public class CGIFGenerator extends SimpleGenerator implements Generator
  {
		/** Name of MarkerTable unit in context. **/
	private static final String MARKER_TABLE_NAME = "MARKER_TABLE";
	
		/**	Type label for negation	relation type. **/
	private	static final String	NEG_TYPE_LABEL = "Neg";

		/**	The newline character for the current platform. **/
	private	static final String	NEW_LINE = System.getProperty("line.separator");

		/**	The indent string. **/
	private	static final String	INDENT = "\t";

		/** Flag indicating whether the generator should allow incomplete relations or
		 * throw an exception. **/
	private boolean allowIncompleteRelations = false;

		/** Flag indicating whether the generator should supress node comments. **/
	private boolean supressNodeComments;

		/** Flag indicating whether the generator should supress graph comments. **/
	private boolean supressGraphComments;

		/** Flag indicating whether the generator should embed concepts which are
		 * only referenced by one relation within that relation. **/
	private boolean embedSingleRelatorConcepts = false;

		/** Flag indicating whether the generator should print a newline before
		 * every graph. **/
	private boolean initiateGraphsWithNewline = false;

		/** Flag indicating whether the generator should print a newline after
		 * every graph. **/
	private boolean terminateGraphsWithNewline = false;

		/** Flag indicating whether the generator should indent nested graphs. **/
	private boolean indentNestedGraphs = false;


    /**
     * Initializes the generator to write to the specified writer
     * using the specified TranslationContext and KnowledgeBase.
     *
     * @param newWriter  the writer to be generated to.
     * @param newKnowledgeBase the knowledge base to be used while generating.
     * @param newTranslationContext  the translationContext to be used while
     * generating.
     */
  public void initializeGenerator(java.io.Writer newWriter,
    KnowledgeBase newKnowledgeBase, 
    TranslationContext newTranslationContext) throws GeneratorException
    {
    super.initializeGenerator(newWriter, newKnowledgeBase, newTranslationContext);
    }

		/** 
		 * Sets a flag indicating whether incomplete relations are allowed whilst generating.
		 * If this flag is false, a GeneratorException will be thrown when an incomplete relation
		 * is encountered.  If it is true, any incomplete arguments will simply be skipped
		 * as if they did not exist.  
		 * If incomplete relations are allowed there is no guarantee that any CGIF parser will
		 * accept the generate output if it enforces relation valence or type definitions.
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 *
		 * @see notio.Relation#isComplete
		 */
  public void setAllowIncompleteRelations(boolean flag)
  	{
  	allowIncompleteRelations = flag;
  	}
    
		/** 
		 * Returns a flag indicating whether incomplete relations are allowed whilst generating.
		 * If this flag is false, a GeneratorException will be thrown when an incomplete relation
		 * is encountered.  If it is true, any incomplete arguments will simply be skipped
		 * as if they did not exist.  
		 * If incomplete relations are allowed there is no guarantee that any CGIF parser will
		 * accept the generate output if it enforces relation valence or type definitions.
		 * The default setting for this flag is false.
		 *
		 * @return  the value for the flag.
		 *
		 * @see notio.Relation#isComplete
		 */
  public boolean getAllowIncompleteRelations()
  	{
  	return allowIncompleteRelations;
  	}
    
		/** 
		 * Sets a flag indicating whether this generator should supress comments in nodes
		 * (Concepts, Relations, and Actors).
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 */
  public void setSupressNodeComments(boolean flag)
  	{
  	supressNodeComments = flag;
  	}
    
		/** 
		 * Returns a flag indicating whether this generator will supress comments in nodes
		 * (Concepts, Relations, and Actors).
		 * The default setting for this flag is false.
		 *
		 * @return the value for the flag.
		 */
  public boolean getSupressNodeComments()
  	{
  	return supressNodeComments;
  	}
    
		/** 
		 * Sets a flag indicating whether this generator should supress comments in graphs.
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 */
  public void setSupressGraphComments(boolean flag)
  	{
  	supressGraphComments = flag;
  	}
    
		/** 
		 * Returns a flag indicating whether this generator will supress comments in graphs.
		 * The default setting for this flag is false.
		 *
		 * @return the value for the flag.
		 */
  public boolean getSupressGraphComments()
  	{
  	return supressGraphComments;
  	}
	
		/** 
		 * Sets a flag indicating whether this generator should generate a newline
		 * at the beginning of every graph.
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 */
  public void setInitiateGraphsWithNewline(boolean flag)
  	{
  	initiateGraphsWithNewline = flag;
  	}
	 
	 	/** 
		 * Returns a flag indicating whether this generator should generate a
		 * newline at the beginning of every graph.
		 * The default setting for this flag is false.
		 *
		 * @return the value for the flag.
		 */
	public boolean getInitiateGraphsWithNewline()
		{
		return initiateGraphsWithNewline;
		}

 		/** 
		 * Sets a flag indicating whether this generator should generate a newline
		 * at the end of every graph.
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 */
  public void setTerminateGraphsWithNewline(boolean flag)
  	{
  	terminateGraphsWithNewline = flag;
  	}
	 
	 	/** 
		 * Returns a flag indicating whether this generator should generate a
		 * newline at the end of every graph.
		 * The default setting for this flag is false.
		 *
		 * @return the value for the flag.
		 */
	public boolean getTerminateGraphsWithNewline()
		{
		return terminateGraphsWithNewline;
		}
 
		/** 
		 * Sets a flag indicating whether this generator should generate
		 * indentation before each graph, increasing the indentation for every
		 * level of nesting.
		 * Note that this will probably not look right unless the generator is set
		 * to "initiate graphs with newline".
		 * The indent will be applied after any newline, intiating or terminating.
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 */
  public void setIndentNestedGraphs(boolean flag)
  	{
  	indentNestedGraphs = flag;
  	}
	 
	 	/** 
		 * Returns a flag indicating whether this generator should generate 
		 * indentation for nested graphs.
		 * The default setting for this flag is false.
		 *
		 * @return the value for the flag.
		 */
	public boolean getIndentNestedGraphs()
		{
		return indentNestedGraphs;
		}


	private String getCurrentIndent(DefiningLabelTable definingLabelTable)
		{
		int indentLevel = definingLabelTable.getCurrentContextLevel();
		StringBuffer buffer = new StringBuffer();

		for (int lev = 0; lev < indentLevel; lev++)
			buffer.append(INDENT);

		return buffer.toString();
		}
		
  	/** 
		 * Sets a flag indicating whether this generator should generate concepts
		 * that are not coreferent to other concepts and only related to a single
		 * relation into the relation itself.
		 * E.g. [Dog *x] (Sleeps ?x) would be rendered as (Sleeps [Dog]).
		 * The default setting for this flag is false.
		 *
		 * @param flag  the new value for the flag.
		 */
  public void setEmbedSingleRelatorConcepts(boolean flag)
  	{
  	embedSingleRelatorConcepts = flag;
  	}
	 
	 	/** 
		 * Returns a flag indicating whether this generator will embed single
		 * relator concepts within the single relation.
		 * The default setting for this flag is false.
		 *
		 * @return the value for the flag.
		 */
	public boolean getEmbedSingleRelatorConcepts()
		{
		return embedSingleRelatorConcepts;
		}
 
 		/**
		 * Returns true if the specified concept is not coreferent with any other
		 * concept is the argument of a single relation only.
		 *
		 * @param concept  the concept to be checked.
		 * @return true if the concept is a single relator concept.
		 */
	private boolean isSingleRelatorConcept(Concept concept)
		{
		CoreferenceSet corefSets[];

		corefSets = concept.getCoreferenceSets();
		if (corefSets.length > 1)
			return false;

		if (corefSets.length == 1 && corefSets[0].size() > 1)
			return false;
			
		return concept.getRelators().length == 1;
		}

	/* Graph Rules */

		/**
     * Generates a vector of graphs to the output stream.
     *
     * @param graphs  the vector of graphs to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     *
     * @bug Does not properly treat translation context.
     */
  public void generateCGStream(Vector graphs) throws GeneratorException
    {
    CGStream(graphs, getDefiningLabelTable(translationContext));
		}

    /**
     * Generates a graph to the output stream.
     *
     * @param graph  the graph to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     *
     * @bug Does not properly treat translation context.
     */
  public void generateOutermostContext(Graph graph) throws GeneratorException
    {
    Graph(graph, getDefiningLabelTable(translationContext));
		}

    /**
     * Generates a graph to the output stream.
     *
     * @param graph  the graph to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  public void generateGraph(Graph graph) throws GeneratorException
    {
    Graph(graph, getDefiningLabelTable(translationContext));
		}

    /**
     * Generates a concept to the output stream.
     *
     * @param concept  the concept to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  public void generateConcept(Concept concept) throws GeneratorException
    {
    Concept(concept, getDefiningLabelTable(translationContext));
    }
    
    /**
     * Generates a relation to the output stream.
     *
     * @param relation  the relation to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  public void generateRelation(Relation relation) throws GeneratorException
    {
    Relation(relation, getDefiningLabelTable(translationContext));
    }
    
    /**
     * Generates an actor to the output stream.
     *
     * @param relation  the relation to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  public void generateActor(Actor actor) throws GeneratorException
    {
    Actor(actor, getDefiningLabelTable(translationContext));
    }

	/* Real generation methods. */    
 
 		/**
     * Generates a CGStream to the output stream.
     *
     * @param graphs  the vector of graphs to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void CGStream(Vector graphs, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
		Enumeration genum;
		Graph gr;

		if (graphs.size() == 0)
			return;

		genum = graphs.elements();

		definingLabelTable.pushContext();

		Graph((Graph)genum.nextElement(), definingLabelTable);

		definingLabelTable.popContext();

		while (genum.hasMoreElements())
			{
			definingLabelTable.pushContext();
			generate(".");
			Graph((Graph)genum.nextElement(), definingLabelTable);
			definingLabelTable.popContext();
			}
		}
    
    /**
     * Generates a graph to the output stream.
     *
     * @param graph  the graph to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void Graph(Graph graph, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
    Concept concepts[];
    Relation relations[];
    String comments[];
    int numConcepts, numRelations, numComments;

		// Push new context onto stack
		definingLabelTable.pushContext();
		
		// Generate an initiating newline if the option is specified
		if (initiateGraphsWithNewline)
			generate(NEW_LINE);
				
		// Generate an indent if the option is specified
		if (indentNestedGraphs)
			generate(getCurrentIndent(definingLabelTable));

    // Generate concepts, defining concepts first, non-defining dominant next,
    // and subordinate last.
    concepts = graph.getConcepts();
    numConcepts = concepts.length;
   
    for (int con = 0; con < numConcepts; con++)
    	{
    	CoreferenceSet corefSets[];
    	
    	corefSets = concepts[con].getCoreferenceSets();
    	
    	for (int set = 0; set < corefSets.length; set++)
    		{
    		try
    			{
    			corefSets[set].setEnableScopeChecking(true);
    			}
    		catch (CorefAddException e)
    			{
    			throw new GeneratorException("Error enabling scope checking in concept.", e);
    			}
    		catch (InvalidDefiningConceptException e)
    			{
    			throw new GeneratorException("Error enabling scope checking in concept.", e);
    			}
    		}
    	}

		if (embedSingleRelatorConcepts)
			{
	    // Generate defining concepts
			for (int con = 0; con < numConcepts; con++)
				if (concepts[con].isDefiningConcept() && !isSingleRelatorConcept(concepts[con]))
					Concept(concepts[con], definingLabelTable);

			// Generate non-defining dominant concepts
			for (int con = 0; con < numConcepts; con++)
				if (!concepts[con].isDefiningConcept() && concepts[con].isDominantConcept() && !isSingleRelatorConcept(concepts[con]))
					Concept(concepts[con], definingLabelTable);

			// Generate subordinate concepts
	    for (int con = 0; con < numConcepts; con++)
  	    if (!concepts[con].isDominantConcept() && !isSingleRelatorConcept(concepts[con]))
    	    Concept(concepts[con], definingLabelTable);
			}
		else
			{
			// Generate defining concepts
			for (int con = 0; con < numConcepts; con++)
				if (concepts[con].isDefiningConcept())
					Concept(concepts[con], definingLabelTable);

			// Generate non-defining dominant concepts
			for (int con = 0; con < numConcepts; con++)
				if (!concepts[con].isDefiningConcept() && concepts[con].isDominantConcept())
					Concept(concepts[con], definingLabelTable);

			// Generate subordinate concepts
	    for (int con = 0; con < numConcepts; con++)
  	    if (!concepts[con].isDominantConcept())
    	    Concept(concepts[con], definingLabelTable);
			}

		// Generate relations and actors.
    relations = graph.getRelations();
    numRelations = relations.length;

    for (int rel = 0; rel < numRelations; rel++)
    	if (relations[rel] instanceof Actor)
	    	Actor((Actor)relations[rel], definingLabelTable);
	    else
	      Relation(relations[rel], definingLabelTable);
		
		// Generate graph comments if not suppressed
		if (!supressGraphComments)
			{
			comments = graph.getComments();
			numComments = comments.length;

			for (int com = 0; com < numComments; com++)
				Comment(comments[com]);
			}

		// Generate a terminating newline if the option is specified
		if (terminateGraphsWithNewline)
			generate(NEW_LINE);
				
		// Pop context from stack
		definingLabelTable.popContext();
		
		// Generate an indent if the option is specified
		if (indentNestedGraphs)
			generate(getCurrentIndent(definingLabelTable));

    }

	/* Concept Rules */
	
    /**
     * Generates a concept to the output stream.
     *
     * @param concept  the concept to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     *
     * @bug Perhaps we should assume a concept was parsed as a literal in a
     * relation if it has no coreference labels and not display it?  Sounds
     * dangerous.  Is there some way to do this intelligently so we don't risk
     * omitting it?  What if it's not a member of a relationship?
     * @bug Removed exception thrown when there is a reference to a coref set that is not
     * in the table by a concept that has multiple sets.  This whole coref thing needs to
     * be reconsidered in light of the idea that single concepts can be generated.
     * @bug Need to review colon generation conditions.
     */
  void Concept(Concept concept, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
    // If a concept is already in the table as a defining concept, write it out with
    //   the corresponding defining label (it's probably a formal parameter).
    // If a concept has no relators or coref sets, just write it out without labels
    // If a concept has relators but no coref sets, create a defining label for it
    //	 as a defining concept and write it out.
    // If a concept has coref sets, but no relators, get the label for the coref set,
    //   or if none exists, create a label and add it for the coref set.  Write out as
    //   bound or defining as appropriate.
    // If a concept has coref sets and relators, treat it as above. ??    
    // Special case: If a concept has one relator and no coref sets, we could write it as
    // a literal once we reach the relator.
    ConceptType conType;
    Referent referent;

		// Grab some objects from the concept
    conType = concept.getType();
    referent = concept.getReferent();

		// Generate concept start
    generate("[");

    // Generate typefield
    if (conType != null)
    	{
    	ConceptType(conType);
    	}

		// Genereate coreference labels
		CoreferenceLabels(concept, definingLabelTable);

    // Generate referent
    Referent(referent, definingLabelTable);

    // Generate concept comment
 		if (!supressNodeComments)
			{
	    Comment(concept.getComment());
	    }
    
    generate("]");
    }

    /**
     * Generates a concept type to the output stream.
     *
     * @param conType  the concept type to be generated.
     * @exception GeneratorException  if an IO error occurs.
     */
  void ConceptType(ConceptType conType) throws GeneratorException
  	{
  	ConceptTypeDefinition typeDef;
  	
    if (conType.getLabel() == null)
    	{
    	typeDef = conType.getTypeDefinition();
    	
    	if (typeDef == null)
    		{
    		// No label and no type definition found.  I guess we can consider it an
    		// 'anonymous and undefined' type and simple leave the field blank.
    		return;
    		}
    	else
    		{
    		// Generate the type definition.
    		ConceptTypeDefinition(typeDef);
    		}
    	}
    else
    	generate(conType.getLabel());
    }

    /**
     * Generates a concept type definition to the output stream.
     *
     * @param typeDef  the concept type definition to be generated.
     * @exception GeneratorException  if an IO error occurs.
     */
  void ConceptTypeDefinition(ConceptTypeDefinition typeDef) throws GeneratorException
  	{
  	Graph differentia;
  	Concept formalParam;
  	ConceptType signature;
  	String corefLabel;
  	DefiningLabelTable definingLabelTable = new DefiningLabelTable();
  
  	generate("(lambda(");
  	
  	// Generate signature
		formalParam = typeDef.getFormalParameter();
		if (formalParam == null)
			{
			signature = typeDef.getSignature();
			if (signature == null)
				throw new GeneratorException("No formal parameters or signature found in type definition.");
			else
				{
				ConceptType(signature);
				corefLabel = definingLabelTable.getNextAvailableDefiningLabel();
				generate("*" + corefLabel);
				}
			}
		else
			{
			signature = formalParam.getType();
			
			// Untyped formal parameters are assumed to use the Universal type.
			if (signature == null)
				signature = knowledgeBase.getConceptTypeHierarchy().getTypeByLabel(ConceptTypeHierarchy.UNIVERSAL_TYPE_LABEL);
				
			ConceptType(signature);
			
			// Output defining label for formal parameter.
			corefLabel = definingLabelTable.getNextAvailableDefiningLabel();

			definingLabelTable.addDefiningConcept(corefLabel, formalParam);
			generate("*" + corefLabel);
			}
			
  	generate(")");
  	
  	// Generate differentia (if any).
  	differentia = typeDef.getDifferentia();
  	if (differentia != null)
  		{
  		Graph(differentia, definingLabelTable);
  		}
  		
  	generate(")");  	
  	}
  	
		/**
     * Generates concept coreference labels to the output stream.
     *
     * @param concept  the concept whose coreference labels are to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void CoreferenceLabels(Concept concept, DefiningLabelTable definingLabelTable) throws GeneratorException
		{
    String corefLabel;
    String corefStuff = "";
    CoreferenceSet corefSets[];
		DefiningLabelTable.TableEntry entry;

		// Generate defining or bound labels (if any)
    // Operates under the assumption that if a coref set is in
    // the table, a defining concept has already been generated for it and so
    // we can make bound label references.

		// If we are embedding single relator concepts, then they don't need
		// coreference labels.
		if (embedSingleRelatorConcepts && isSingleRelatorConcept(concept))
			return;

		// Look for the concept is in the table.
		entry = definingLabelTable.findEntryByConcept(concept);

		// If the concept is in the table as a defining concept, then simply write out
		// the appropriate defining label.  
		// This should only happen if we are generating concepts directly,
		// instead of as part of a graph, and we generate the same concept twice.
		// The other case is when we are generating a type definition and the
		// label was added for use by a formal parameter.
		if (entry != null && entry.isDefiningEntry)
			{
    	corefStuff += "*" + entry.label;

			generate(corefStuff);
			return;
			}

		// If the concept is in the table but is not a defining concept,
		// then write it out as a bound label.
		// This should only happen if we are generating concepts directly,
		// instead of as part of a graph, and we generate the same concept twice.
		// We could also a preference for whether re-generated concepts are
		// generated using bound or defining labels.
		if (entry != null && !entry.isDefiningEntry)
			{
			corefLabel = entry.label;
    	corefStuff += "?" + corefLabel;
			generate(corefStuff);
			return;
			}

		// ** The concept is not in the table. **

		// Check to see if it has any coreference sets.
		corefSets = concept.getCoreferenceSets();

   	// If it has zero coref sets, but has one or more relators, create a defining 
   	// label for it and establish it as a defining concept in the table.
   	if (corefSets == null || corefSets.length == 0)
     	{
     	Relation relators[];
     
   		// Check to see if it has one or more relators before bothering to create a label 
			// for it.  If it has no relators, it does not need a label.
     	relators = concept.getRelators();
     	if ((relators != null) && (relators.length > 0))
     		{      
				corefLabel = definingLabelTable.getNextAvailableDefiningLabel();

				definingLabelTable.addDefiningConcept(corefLabel, concept);
				corefStuff += "*" + corefLabel;
				}
			
			generate(corefStuff);
			return;
     	} 
		
		// If it has one or more coreference sets...
    for (int setCount = 0; setCount <  corefSets.length; setCount++)
			{
	   	if (corefSets.length == 1)      
				{
				// Look for coreference set in the table.
				entry = definingLabelTable.findEntryByCoreferenceSet(corefSets[setCount]);
	
				// If the coreference set is in the table use the label to make a reference.
				if (entry != null)
					{
						// If the concept is the defining concept of the coref set, make a
						// defining label, otherwise make a bound label.
						if (concept.isDefiningConcept())
		  	     	corefStuff += "*" + entry.label;
						else
		  	     	corefStuff += "?" + entry.label;
	
					generate(corefStuff);
					return;
					}
	
	
	     	// The coref set is not in the table so search using the defining
				// concept, just in case that was added and the coref set wasn't.
				entry = definingLabelTable.findEntryByDefiningConcept(corefSets[setCount].getDefiningConcept());
	
				if (entry != null)
					{
						// Update the table so to include the coreference set.
					definingLabelTable.addCoreferenceSet(entry.label, corefSets[setCount]);
						// Add this concept as a bound concept to the table.
						// Note that if this concept has multiple coref sets, it will be
						// associated with the label of the last set only.
					definingLabelTable.addBoundConcept(entry.label, concept);
						// Generate bound label.
					corefStuff += "?" + entry.label;
	
					generate(corefStuff);
					return;
					}
	
				// Neither the coreference set nor its defining concept are in the
				// table so we need to add a label.
				
				// If the coreference set suggests a label, use if if it isn't already
				// in use, otherwise, create a new label.
				corefLabel = corefSets[setCount].getDefiningLabel();
				if (corefLabel == null || definingLabelTable.findFirstEntryByLabel(corefLabel) != null)
	        corefLabel = definingLabelTable.getNextAvailableDefiningLabel();
	
				// Add coreference set (and its defining concept) to the table under the
				// new label.
				definingLabelTable.addCoreferenceSet(corefLabel, corefSets[setCount]);
	
				// If the concept is a defining concept, generate a defining label,
				// otherwise generate a bound label.
				if (concept.isDefiningConcept())
	       	corefStuff += "*" + corefLabel;
				else
					{
	       	corefStuff += "?" + corefLabel;
						// Add this concept as a bound concept to the table.
						// Note that if this concept has multiple coref sets, it will be
						// associated with the label of the last set only.
					definingLabelTable.addBoundConcept(entry.label, concept);
					}
	
				generate(corefStuff);
				return;
				}
			}
		}
 
    /**
     * Generates a referent to the output stream.
     *
     * @param referent  the referent to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an IO error occurs.
     */
  void Referent(Referent referent, DefiningLabelTable definingLabelTable) throws GeneratorException
  	{
  	Graph descriptor;
  	
  	if (referent == null)
  		return;
  	
		generate(":");

 		Quantifier(referent.getQuantifier());
		Designator(referent.getDesignator(), definingLabelTable);
		
		descriptor = referent.getDescriptor();
		if (descriptor != null)
			Graph(descriptor, definingLabelTable);
  	}
  	
    /**
     * Generates a quantifier to the output stream.
     *
     * @param quantifier  the quantifier to be generated.
     * @exception GeneratorException  if an IO error occurs.
     */
   void Quantifier(Macro quantifier) throws GeneratorException
    {
    if (quantifier != null)
      generate(quantifier.getName() + " ");
    }

    /**
     * Generates a designator to the output stream.
     *
     * @param designator  the designator to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void Designator(Designator designator, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
    if (designator == null)
      return;

    switch (designator.getDesignatorKind())
      {
      case Designator.DESIGNATOR_LITERAL:
      	{
				LiteralDesignator((LiteralDesignator)designator);
        break;
        }

      case Designator.DESIGNATOR_MARKER:
        {
				MarkerDesignator((MarkerDesignator)designator);
        break;
        }

      case Designator.DESIGNATOR_NAME:
        {
				NameDesignator((NameDesignator)designator);
        break;
        }

      case Designator.DESIGNATOR_DEFINED:
       	throw new GeneratorException("Defined designators not handled by this generator.");
        
      default:
       	throw new GeneratorException("Unknown designator kind encountered in concept.");
       	
      }
    }

    /**
     * Generates a literal designator to output stream.
     *
     * @param designator  the designator to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void LiteralDesignator(LiteralDesignator designator) throws GeneratorException
  	{
  	Object literal;
      	
   	literal = designator.getLiteral();
   	if (literal instanceof String)
       generate("\"" + escapeCharactersInString(literal.toString(), "\"", '\"') + "\"");
    else if (literal instanceof Number)
       generate(" " + literal.toString());
    else
      throw new GeneratorException("Unknown type of literal designator found in concept.");
  	}
    
    /**
     * Generates a marker designator to output stream.
     *
     * @param designator  the designator to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void MarkerDesignator(MarkerDesignator designator) throws GeneratorException
    {
    Marker marker;
    String markerID;

    marker = designator.getMarker();
    generate("#" + marker.getMarkerID());
    }
    
    /**
     * Generates a name designator to output stream.
     *
     * @param designator  the designator to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void NameDesignator(NameDesignator designator) throws GeneratorException
    {
    String name;

    name = designator.getName();
    generate("'" + escapeCharactersInString(name, "'", '\'') + "'");
    }
    
    /**
     * Generates a comment to output stream.
     *
     * @param comment  the comment to be generated.
     * @exception GeneratorException  if an error occurs while generating.
     */
  void Comment(String comment) throws GeneratorException
    {
    if (comment == null)
      return;

    if (comment.length() == 0)
      return;

    generate(";"+escapeCharactersInString(comment, ";", ';') +";");
    }

	/* Relation Rules */

    /**
     * Generates a relation to the output stream.
     *
     * @param relation  the relation to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     *
     * @bug Literal concept handling is just plain wrong.
     * @bug Removed throwing of exceptions when there is no coref label since
     * just doesn't make sense if we want to print a single relation.
     * @bug Can relations not have a blank type field?
     */
  void Relation(Relation relation, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
    RelationType relType;
    Concept arguments[];

		// Generate relation start
    generate("(");
   
    // Generate typefield
    relType = relation.getType();
    if (relType != null)
    	RelationType(relType);
    else
      throw new GeneratorException("No type found while generating relation.");

    // Generate arguments as bound labels
    arguments = relation.getArguments();

    for (int arg = 0; arg < arguments.length; arg++)
      Arc(arguments[arg], definingLabelTable);

		// Generate relation comment
 		if (!supressNodeComments)
	    Comment(relation.getComment());

		// Generate relation terminator
    generate(")");
    }

    /**
     * Generates a relation type to the output stream.
     *
     * @param relType  the relation type to be generated.
     * @exception GeneratorException  if an IO error occurs.
     *
     * @bug What about types with no label or definition?  Anonymous types?
     * @bug What about types with both a label and a definition?
     */
  void RelationType(RelationType relType) throws GeneratorException
  	{
  	RelationTypeDefinition typeDef;
    if (relType.getLabel() == null)
    	{
    	// Generate type definition here instead as a lambda expression
    	typeDef = relType.getTypeDefinition();
    	
    	if (typeDef != null)
    		RelationTypeDefinition(typeDef);
    	else
    		throw new GeneratorException("Encountered relation type with no label or definition.");
    	}
    else
    	generate(relType.getLabel());
    }

    /**
     * Generates a relation type definition to the output stream.
     *
     * @param typeDef  the relation type definition to be generated.
     * @exception GeneratorException  if an IO error occurs.
     */
  void RelationTypeDefinition(RelationTypeDefinition typeDef) throws GeneratorException
  	{
  	Graph differentia;
  	Concept formalParams[];
  	ConceptType signature[];
  	String corefLabel;
  	DefiningLabelTable definingLabelTable = new DefiningLabelTable();
  	
  	generate("(lambda(");
  	
  	// Generate signature
		formalParams = typeDef.getFormalParameters();
		if (formalParams == null)
			{
			signature = typeDef.getSignature();
			if (signature == null)
				throw new GeneratorException("No formal parameters or signature found in type definition.");
			else
				{
				for (int sig = 0; sig < signature.length; sig++)
					{
					if (sig > 0)
						generate(",");
						
					ConceptType(signature[sig]);
					corefLabel = definingLabelTable.getNextAvailableDefiningLabel();
					generate("*" + corefLabel);
					}
				}
			}
		else
			{
			for (int sig = 0; sig < formalParams.length; sig++)
				{
				ConceptType sigType;
				
				if (sig > 0)
					generate(",");

				sigType = formalParams[sig].getType();
			
				// Untyped formal parameters are assumed to use the Universal type.
				if (sigType == null)
					sigType = knowledgeBase.getConceptTypeHierarchy().getTypeByLabel(ConceptTypeHierarchy.UNIVERSAL_TYPE_LABEL);
				
				ConceptType(sigType);
			
				// Output defining label for formal parameter.
				corefLabel = definingLabelTable.getNextAvailableDefiningLabel();

				definingLabelTable.addDefiningConcept(corefLabel, formalParams[sig]);
				generate("*" + corefLabel);
				}
			}
			
  	generate(")");
  	
  	// Generate relator (if any).
  	differentia = typeDef.getRelator();
  	if (differentia != null)
  		{
  		Graph(differentia, definingLabelTable);
  		}
  		
  	generate(")");  	
  	}

    /**
     * Generates an arc to the output stream.
     *
     * @param argument  the relation argument to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     */
  final void Arc(Concept argument, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
    // Check for null argument (incomplete relation)
    if (argument == null)
      {
      if (!allowIncompleteRelations)
      	throw new GeneratorException("Incomplete relation encountered during generation.");
      // If we are allowing incomplete relations then we simply generate nothing for 
      // this argument.
      }
    else
      {
      CoreferenceSet corefSets[];
			DefiningLabelTable.TableEntry entry;

      // Check if this concept is in the table.
			entry = definingLabelTable.findEntryByConcept(argument);

			if (entry == null)
      	{
				// Concept is not in the table.
      	// Check if it has coreference sets
	      corefSets = argument.getCoreferenceSets();

	      if (corefSets == null || corefSets.length == 0)
	        {
    	    // This should never happen unless we are generating less than the entire graph.  
    	    // Otherwise we would have generated the concept earlier, discovered this relator,
    	    // and added it to the table as a defining concept.
    	    // Since this is possible, we might as well write a literal here.  
    	    // Let's do that for now.
    	    // The other case where this can occur is if we allow use of literal concepts
    	    // in relators which exclusively relate said concepts.  In this case, we would
    	    // have deliberately skipped generating the concept, leading us to this point.
    	    // Possibly dangerous if we allow defining label tables to be imported prior
    	    // to generation.
		      Concept(argument, definingLabelTable);
	        }
  	    else
    	    {
      	  entry = definingLabelTable.findEntryByCoreferenceSet(corefSets[0]);
	        if (entry == null)
    	      {
  	        // Again, this should never happen unless we are generating less than the
  	        // entire graph.  See above.
  	  	    Concept(argument, definingLabelTable);
          	}
        	else
          	generate("?" + entry.label);
        	}
        }
      else
      	{
       	generate("?" + entry.label);
      	}
      }
    }
  	
	/* Actor Rules */
	
    /**
     * Generates a actor to the output stream.
     *
     * @param actor  the actor to be generated.
		 * @param definingLabelTable  the current defining label table.
     * @exception GeneratorException  if an error occurs while generating.
     *
     * @bug Literal concept handling is just plain wrong.
     * @bug Must be entirely reexamined and updated based on Relation() method. 
     * (e.g. allowIncompleteRelation)
     */
  void Actor(Actor actor, DefiningLabelTable definingLabelTable) throws GeneratorException
    {
    RelationType relType;
    Concept inArcs[];
    Concept outArcs[];

		// Generate actor start
    generate("<");
   
    // Generate typefield
    relType = actor.getType();
    if (relType != null)
    	RelationType(relType);
    else
      throw new GeneratorException("No type found while generating actor.");

    // Generate input arguments as bound labels
    inArcs = actor.getInputArguments();

    for (int arg = 0; arg < inArcs.length; arg++)
      Arc(inArcs[arg], definingLabelTable);

		// Generate input/output arc delimiter
		generate("|");

    // Generate output arguments as bound labels
    outArcs = actor.getOutputArguments();

    for (int arg = 0; arg < outArcs.length; arg++)
      Arc(outArcs[arg], definingLabelTable);
      
		// Generate actor comment
 		if (!supressNodeComments)
	  	Comment(actor.getComment());

		// Generate actor terminator
    generate(">");
    }
  }
