package notio.translators;

import java.util.*;
import java.io.Serializable;
import notio.*;

    /** 
     * A TranslationInfoUnit that serves as a symbol table relating defining labels to
     * defining concepts and coreference sets.  This table allows stack-like operations
     * for maintaining label scope.
     *
     * @author Finnegan Southey
     * @version $Name:  $ $Revision: 1.12 $, $Date: 2001/07/03 10:33:19 $
		 * @legal Copyright (c) Finnegan Southey, 1996-1999
		 *	This program is free software; you can redistribute it and/or modify it 
		 *	under the terms of the GNU Library General Public License as published 
		 *	by the Free Software Foundation; either version 2 of the License, or 
		 *	(at your option) any later version.  This program is distributed in the 
		 *	hope that it will be useful, but WITHOUT ANY WARRANTY; without even the 
		 *	implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
		 *	See the GNU Library General Public License for more details.  You should 
		 *	have received a copy of the GNU Library General Public License along 
		 *	with this program; if not, write to the Free Software Foundation, Inc., 
		 *	675 Mass Ave, Cambridge, MA 02139, USA.
		 *
		 * @idea Consider extending "pop" operations to the label generator.
     */
public class DefiningLabelTable extends SimpleInfoUnit
  {
		/** A flag that indicates whether debugging messages should be shown for
		 * push/pop operations on the context stack. **/
	private static final boolean SHOW_STACK_OPS = false;
 
	  /** A factory for producing defining labels. **/
  private LabelFactory defLabelFactory;
  
		/** The base table of the ContextTable stack. **/
	private ContextTable baseTable;
	
		/** The top table of the ContextTable stack. **/
	private ContextTable topTable;
	
		/**
		 * An inner class used to hold table entries.
		 */
	public final class TableEntry
		{
			/** A flag indicating whether this a defining entry or not. **/
		boolean isDefiningEntry;

			/** The label of this entry. **/
		String label;
		
			/** The concept of this entry. **/
		Concept concept;
		
			/** The coreference set of this entry. **/
		CoreferenceSet corefSet;
		
			/** The context table to which this entry belongs. **/
		ContextTable contextTable;

			/** The entry before this entry in the current context. **/
		TableEntry prevEntry;
		
			/**
			 * Constructs a new TableEntry with the specified values.
			 *
			 * @param newLabel  the label for this entry.
			 * @param newConcept  the concept for this entry.
			 * @param newDefiningFlag  a flag indicating whether this is a defining
			 * concept or not.
			 * @param newConTable  the context table to which the entry belongs.
			 */
		TableEntry(String newLabel, Concept newConcept, boolean newDefiningFlag, ContextTable newConTable)
			{
			label = newLabel;
			concept = newConcept;
			isDefiningEntry = newDefiningFlag;
			contextTable = newConTable;
			}

			/**
			 * Returns the context level to which this entry belongs.
			 *
			 * @return the context level to which this entry belongs.
			 */
		public final int getContextLevel()
			{
			return contextTable.getContextLevel();
			}
		}
	
		/**
		 * An inner class used to build a stack of sub-tables.
		 */
	private class ContextTable extends Hashtable
		{
			/** The context level of this subtable. **/
		int contextLevel;
		
			/** The ContextTable that is parent to this one. **/
		ContextTable parent;
		
		  /** A lookup table that maps coreference labels to TableEntries. **/
		Hashtable defLabelToEntryTable;

		  /** A lookup table that maps coreference sets to TableEntries. **/
		Hashtable corefSetToEntryTable;
  
		  /** A lookup table that maps defining concepts to TableEntries. **/
		Hashtable conceptToEntryTable;
		
			/**
			 * Constructs a new ContextTable that serves as a base for other tables.
			 */
		ContextTable()
			{
			defLabelToEntryTable = new Hashtable();
			corefSetToEntryTable = new Hashtable();
			conceptToEntryTable = new Hashtable();
			
			contextLevel = 0;
			parent = null;
			}
			
			/**
			 * Constructs a new ContextTable with the specified table as a base.
			 *
			 * @param parentTable  the base table for this table.
			 */
		ContextTable(ContextTable parentTable)
			{
			this();
			parent = parentTable;
			contextLevel = parent.getContextLevel() + 1;
			}
			
			/**
			 * Returns the context level of this subtable.
			 *
			 * @return the context level of this subtable.
			 */
		int getContextLevel()
			{
			return contextLevel;
			}
		
			/**
			 * Returns the parent of this subtable.
			 *
			 * @return the parent of this subtable.
			 */
		ContextTable getParent()
			{
			return parent;
			}		
			
			/**
			 * Returns a deep copy of this ContextTable, such that modifying the
			 * copy does not affect the original.
			 *
			 * @return a copy of this ContextTable.
			 */
		ContextTable copy()
			{
			ContextTable newTable = new ContextTable();
			
			newTable.parent = parent;
			newTable.contextLevel = contextLevel;
			newTable.defLabelToEntryTable = copyHashtable(defLabelToEntryTable);
			newTable.corefSetToEntryTable = copyHashtable(corefSetToEntryTable);
			newTable.conceptToEntryTable = copyHashtable(conceptToEntryTable);
			
			return newTable;
			}
		}
		
    /**
     * Constructs a new DefiningLabelTable.
     */
  public DefiningLabelTable()
    {
		baseTable = new ContextTable();
		topTable = baseTable;
		defLabelFactory = new LabelFactory();
    }

    /**
     * Constructs a new DefiningLabelTable by copying the specified table.
		 *
		 * @param orgTable  the table to be copied.
     */
  public DefiningLabelTable(DefiningLabelTable orgTable)
    {
		ContextTable currentOrgTable = orgTable.getTopTable();
		ContextTable currentNewTable;
		
		topTable = currentOrgTable.copy();
		currentNewTable = topTable;
		currentOrgTable = currentOrgTable.parent;
		
		
		while (currentOrgTable != null)
			{				
			currentNewTable.parent = currentOrgTable.copy();
			currentNewTable = currentNewTable.parent;
			currentOrgTable = currentOrgTable.parent;
			}
		
		baseTable = currentNewTable;
		defLabelFactory = new LabelFactory(orgTable.getLastDefiningLabel());
    }

		/**
		 * Pushes a new context onto the stack.
		 * Any entries added to this table after the push will be removed when a corresponding
		 * pop is called.
		 */
	public final void pushContext()
		{
		if (SHOW_STACK_OPS)
			System.out.println("PUSH");

			// Add a new Subtable to the stack.
		topTable = new ContextTable(topTable);
		}

		/**
		 * Pops a context off the stack.
		 * Any entries added to this table since the corresponding the push will be removed.
		 */
	public final void popContext()
		{
		if (SHOW_STACK_OPS)
			System.out.println("POP");

			// Remove the top subtable from the stack.
		topTable = topTable.getParent();
		if (topTable == null)
			{
			System.err.println("Fatal error: Popped one too many contexts from a DefiningLabelTable.");
			System.exit(1);
			}
		}

		/**
		 * Finds the entry corresponding to the specified Concept.
		 *
		 * @param concept  the concept to be found.
		 * @return the entry, or null.
		 */
	public final TableEntry findEntryByConcept(Concept concept)
		{
		ContextTable currentTable = topTable;
		TableEntry entry;
		
		while (currentTable != null)
			{
			entry = (TableEntry)currentTable.conceptToEntryTable.get(concept);
			
			if (entry != null)
				return entry;
			
			currentTable = currentTable.getParent();
			}
		
		return null;
		}
	
		/**
		 * Finds the entry corresponding to the specified defining Concept.
		 * This returns null if the concept is not in the table or is not a
		 * defining concept.
		 *
		 * @param defConcept  the defining concept to be found.
		 * @return the entry, or null.
		 */
	public final TableEntry findEntryByDefiningConcept(Concept defConcept)
		{
		TableEntry entry;

		entry = findEntryByConcept(defConcept);
		
		if (entry != null && entry.isDefiningEntry)
			return entry;
		else
			return null;
		}
		
		/**
		 * Finds the entry corresponding to the specified CoreferenceSet.
		 *
		 * @param corefSet  the coreference set to be found.
		 * @return the entry, or null.
		 */
	public final TableEntry findEntryByCoreferenceSet(CoreferenceSet corefSet)
		{
		ContextTable currentTable = topTable;
		TableEntry entry;
		
		while (currentTable != null)
			{
			entry = (TableEntry)currentTable.corefSetToEntryTable.get(corefSet);
			
			if (entry != null)
				return entry;
			
			currentTable = currentTable.getParent();
			}
		
		return null;
		}
		
		/**
		 * Removes all entries corresponding to the specified label.
		 * If no entry is associated with the label, nothing happens.
		 *
		 * @param label  the label to be removed.
		 */
	private final void removeEntryByLabel(String label)
		{
		TableEntry entry;
	
		while ((entry = findFirstEntryByLabel(label)) != null)
			{
			entry.contextTable.defLabelToEntryTable.remove(label);
			entry.contextTable.corefSetToEntryTable.remove(entry.corefSet);
			entry.contextTable.conceptToEntryTable.remove(entry.concept);
			}
		}

		/**
		 * Finds an entry in the current context corresponding to the specified label.
		 *
		 * @param label  the label to be found.
		 * @return the entry, or null.
		 */
	public final TableEntry findLocalEntryByLabel(String label)
		{
		return (TableEntry)topTable.defLabelToEntryTable.get(label);
		}
	
	
		/**
		 * Finds the first (topmost) defining entry corresponding to the specified label.
		 *
		 * @param label  the label to be found.
		 * @return the entry, or null.
		 */
	public final TableEntry findDefiningEntryByLabel(String label)
		{
		ContextTable currentTable = topTable;
		TableEntry entry;
		
		while (currentTable != null)
			{
			entry = (TableEntry)currentTable.defLabelToEntryTable.get(label);

				// Search back along the chain of entries.
			while (entry != null && !entry.isDefiningEntry)
			{
				entry = entry.prevEntry;
			}

			if (entry != null && entry.isDefiningEntry)
				return entry;
			
			currentTable = currentTable.getParent();
			}
		
		return null;
		}
	
		/**
		 * Finds the first (topmost) entry corresponding to the specified label.
		 *
		 * @param label  the label to be found.
		 * @return the entry, or null.
		 */
	public final TableEntry findFirstEntryByLabel(String label)
		{
		ContextTable currentTable = topTable;
		TableEntry entry;
		
		while (currentTable != null)
			{
			entry = (TableEntry)currentTable.defLabelToEntryTable.get(label);
			
			if (entry != null)
				return entry;
			
			currentTable = currentTable.getParent();
			}
		
		return null;
		}
	
		/**
		 * Adds an entry to the table that relates the specified label to the
		 * specified concept with a flag indicating whether it is a defining or
		 * bound concept.
		 * If the concept is a defining concept with an associated coreference
		 * set, that set will also be included in the entry.
		 *
		 * @param newLabel  the label to be be added.
		 * @param newConcept  the defining concept to be added.
		 * @param defFlag  a flag indicating whether the concept is defining or
		 * not.
		 *
		 * @exception java.lang.IllegalArgumentException  if the concept is
		 * flagged as a defining concept but it is a member of multiple
		 * coreference sets.
		 */
	private final void addConcept(String newLabel, Concept newConcept, boolean defFlag)
		{
		TableEntry newEntry;
		TableEntry prevEntry;

		newEntry = new TableEntry(newLabel, newConcept, defFlag, topTable);

			// Link to previous entry in the current context (if any).
		prevEntry = findLocalEntryByLabel(newLabel);

		if (prevEntry != null)
			newEntry.prevEntry = prevEntry;

			// Add new entry to lookup tables (replaces old label entries
			// automatically).
		topTable.defLabelToEntryTable.put(newLabel, newEntry);
		if (newConcept == null)
			System.err.println("GASFASFD\n");
		topTable.conceptToEntryTable.put(newConcept, newEntry);

			// If we're adding a defining concept, check for a coreference set and
			// put it in the entry if it exists.
		if (defFlag)
			{
			CoreferenceSet corefSets[];

			corefSets = newConcept.getCoreferenceSets();
			if (corefSets != null && corefSets.length > 0)
				if (corefSets.length == 1)
					{
						newEntry.corefSet = corefSets[0];
						newEntry.contextTable.corefSetToEntryTable.put(newEntry.corefSet, newEntry);
					}
				else
					{
						// This is an error, since defining concepts shouldn't have more
						// than one coreference set.
						// Throw an exception.
						throw new IllegalArgumentException("Attempted to add defining concept with multiple coreference sets.");
					}
			}
		}

		/**
		 * Adds an entry to the table that relates the specified label to the
		 * specified defining concept.
		 * If the defining concept has an associated coreference set, that set 
		 * will also be included in the entry.
		 *
		 * @param newLabel  the label to be be added.
		 * @param newDefConcept  the defining concept to be added.
		 */
	public final void addDefiningConcept(String newLabel, Concept newDefConcept)
		{
		addConcept(newLabel, newDefConcept, true);
		}
	
		/**
		 * Adds an entry to the table that relates the specified label to the
		 * specified bound concept.
		 *
		 * @param newLabel  the label to be be added.
		 * @param newBoundConcept  the bound concept to be added.
		 */
	public final void addBoundConcept(String newLabel, Concept newBoundConcept)
		{
		addConcept(newLabel, newBoundConcept, false);
		}

    /**
		 * Adds an entry to the table that relates the specified label to the
		 * specified coreference set.
		 * If the coreference set is already associated with the label, nothing will done.
		 * If the coreference set has an associated defining concept, that concept
		 * will also be included in the entry.
     *
     * @param definingLabel  the defining label being mapped.
     * @param corefSet  the coreference set being mapped.
		 *
		 * @exception java.lang.IllegalArgumentException  if the label is already
		 * associated with some other coreference set,
		 * or if it is associated with a defining concept other than the one in
		 * the coreference set.
		 *
		 * @note What about allowing "hiding"?
		 * @bug Should this be setting "isDefiningEntry" to true or not?
     */
  public void addCoreferenceSet(String definingLabel, CoreferenceSet corefSet)
    {
		TableEntry entry;
		Concept defConcept;
		
		entry = findDefiningEntryByLabel(definingLabel);
	
			// If an entry already exists and has a coreference set...
		if (entry != null && entry.corefSet != null)
			{
				// The entry contains the set being added, do nothing and return.
				// Otherwise, throw an IllegalArgumentException.
			if (entry.corefSet.equals(corefSet))
				return;
			else
				throw new IllegalArgumentException("Attempted to associate a coreference set with a label already associated with a coreference set.");
			}
	
			// If the coref set has a defining concept different from that already
			// in the entry, throw an IllegalArgumentException.
		defConcept = corefSet.getDefiningConcept();
		if (entry != null && defConcept != null && !entry.concept.equals(defConcept))
		{
			throw new IllegalArgumentException("Attempted to associate a coreference set with a label already associated with defining concept from a different set.");
		}

		if (entry == null)
			{
			entry = new TableEntry(definingLabel, null, false, topTable);
			topTable.defLabelToEntryTable.put(definingLabel, entry);
			}

		entry.isDefiningEntry = true;
		entry.corefSet = corefSet;
		entry.contextTable.corefSetToEntryTable.put(corefSet, entry);
		
		if (defConcept != null)
			{
			entry.concept = defConcept;
			entry.contextTable.conceptToEntryTable.put(defConcept, entry);
			}
    }
 

    /**
     * Clears all existing mappings between coreference labels and coreference sets.
     */
  public void clearDefiningLabelToCoreferenceSetMapping()
    {
		ContextTable currentTable = topTable;
		
		while (currentTable != null)
			{
			Enumeration elements;
		
			elements = currentTable.corefSetToEntryTable.elements();
		
			while (elements.hasMoreElements())
				{
				TableEntry entry;

				entry = (TableEntry)elements.nextElement();
				
				currentTable.corefSetToEntryTable.remove(entry.corefSet);
				entry.corefSet = null;
				if (entry.concept == null)
					currentTable.defLabelToEntryTable.remove(entry.label);
					
				}
						
			currentTable = currentTable.getParent();
			}
    }

    /**
     * Clears all existing mappings between defining concepts and defining labels.
     */
  public void clearDefiningLabelToDefiningConceptMapping()
    {
		ContextTable currentTable = topTable;
		
		while (currentTable != null)
			{
			Enumeration elements;
		
			elements = currentTable.conceptToEntryTable.elements();
		
			while (elements.hasMoreElements())
				{
				TableEntry entry;

				entry = (TableEntry)elements.nextElement();
				
				currentTable.conceptToEntryTable.remove(entry.concept);
				entry.concept = null;
				if (entry.corefSet == null)
					currentTable.corefSetToEntryTable.remove(entry.corefSet);					
				}
			
			currentTable = currentTable.getParent();
			}    
		}
    
    /**
     * Returns a new defining label that is not already present in the defining 
     * label/coreference set mapping.  There are no constraints on the form of the
     * label other than that they conform to a valid CGIF identifier.  Otherwise,
     * labels' forms are determined by the underlying implementation.
     *
     * @return a new, unique, and unused defining label.
     */
  public String getNextAvailableDefiningLabel()
    {
    String newLabel;
    
    do
    	{
    	newLabel = defLabelFactory.getNextLabel();
    	}
    while (findFirstEntryByLabel(newLabel) != null);
      
    return newLabel;
    }
    
    /**
     * Resets the generator for defining labels so that it starts again.
     * This method would typically be called after clearing the defining label - 
     * coreference set mapping so that labels may be reused in a new translation, 
     * if possible.
     * Implementations are not required to honour this although they should silently
     * do nothing rather than throwing an exception.  If possible, implementations
     * should reset their label generators so that they produce a sequence of
     * labels as if this were a newly constructed context.
     */
  public void resetAvailableDefiningLabel()
    {
    defLabelFactory = new LabelFactory();
    }
    
		
		/**
		 * Returns the current context level of this table.
		 * This is the number of contexts sitting on the stack minus one.
		 * It is never less than 0.
		 *
		 * @return the current context level of this table.
		 */
	public int getCurrentContextLevel()
		{
			return topTable.getContextLevel();
		}
		
    /**
     * Returns the top ContextTable of this table.
     *
     * @return the top ContextTable of this table.
     */
  ContextTable getTopTable()
  	{
  	return topTable;
  	}

    /**
     * Returns the last defining label generated by this context or null if no labels
     * have generated.
     * Note that this method does not generated any new labels.
     *
     * @return the last defining label generated by this context or null.
     */
  public String getLastDefiningLabel()
  	{
  	return defLabelFactory.getLastLabel();
  	}
  	  	
		/**
		 * Creates a deep copy of the specified hashtable.
		 *
		 * @param inTable  the hashtable to be copied.
		 * @return a copy of the hashtable.
		 */
	private static Hashtable copyHashtable(Hashtable inTable)
		{
		Enumeration keys, elements;
		Hashtable newTable = new Hashtable();
		
		keys = inTable.keys();
		elements = inTable.elements();
		
		while (keys.hasMoreElements())
			newTable.put(keys.nextElement(), elements.nextElement());
			
		return newTable;
		}

	  /**
  	 * Resets this information unit to its initial state.  This method can be called
  	 * by translators in order to ensure that a unit contains no information relating
  	 * to earlier translation sessions.
		 * This causes all contexts to be emptied and popped and fresh initial context pushed.
  	 */
  public void resetUnit()
  	{
		resetAvailableDefiningLabel();
		baseTable= new ContextTable();
		topTable = baseTable;
		}

  	/**
  	 * Returns a duplicate of this information unit that is distinct from the original.
  	 * This means that changes to the original will not affect the duplicate and vice
  	 * versa.
  	 *
  	 * @return a duplicate of this unit.
  	 */
  public TranslationInfoUnit copyUnit()
  	{
  	return new DefiningLabelTable(this);
  	}
  }
